# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-11-24 20:24
from __future__ import unicode_literals

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BadgeAward',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.CharField(max_length=255, verbose_name='slug')),
                ('level', models.IntegerField(verbose_name='level')),
                ('points', models.PositiveIntegerField(default=0, verbose_name='points')),
                ('awarded_on', models.DateTimeField(default=datetime.datetime.now, verbose_name='awarded on')),
                ('created_on', models.DateTimeField(auto_now_add=True, verbose_name='created on')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='badges_earned', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'badge award',
                'verbose_name_plural': 'badge awards',
            },
        ),
    ]



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import BadgeAward


@admin.register(BadgeAward)
class BadgeAwardAdmin(admin.ModelAdmin):
    THUMB_SIZE = 40

    list_display = ('user', 'slug', 'name', 'level', 'points', 'awarded_on', 'image')
    raw_id_fields = ('user', )
    list_filter = ('slug', 'level', 'points', )
    search_fields = ('slug', 'level')

    def image(self, obj):
        return '<img src="{path}" width="{size}" height="{size}">'.format(
            path=obj.image_url, size=self.THUMB_SIZE)
    image.allow_tags = True
    image.short_description = _('image')



# coding: utf-8
from __future__ import unicode_literals

from django import template
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.db.models import Count
from django.utils.functional import SimpleLazyObject
from django.contrib.auth import get_user_model

from ..models import BadgeAward
from ..internals import badges

User = get_user_model()
register = template.Library()


@register.assignment_tag
def badges_for_user(user):
    """
    Usage:

        {% load badges_tags %}

        {% badges_for_user user as badges %}

        {% for badge in badges %}
            Name: {{badge.name}}
            ...
        {% endfor%}
    """

    return BadgeAward.objects \
        .filter(user=user) \
        .order_by("-awarded_on")


@register.assignment_tag
def all_badges_with_awarded_flag_for(user):
    user_badges = badges_for_user(user)
    badges_awarded = {(b.slug, b.level): b for b in user_badges}
    all_badges = []
    for slug, c in badges._registry.items():
        for level in range(len(c.levels)):
            badge = BadgeAward(
                user=user,
                slug=slug,
                level=level
            )
            key = (slug, level)
            badge.awarded = key in badges_awarded
            if badge.awarded:
                badge.awarded_on = badges_awarded[key].awarded_on
            all_badges.append(badge)
    return all_badges


class MostAwardedBadge(object):

    def __init__(self, detail, count, slug=None, users_count=None):
        self.detail = detail
        self.slug = getattr(detail, 'slug', slug)
        self.count = count
        self.users_count = users_count
        self.pct = (count / float(users_count) * 100) if users_count else 100

    def __getattr__(self, attr):
        return getattr(self.detail, attr)

    def __repr__(self):
        return "MostAwardedBadge(detail={detail!r}, count={count})".format(
            detail=self.detail,
            count=self.count,
        )

    @property
    def image_url(self):
        slug = getattr(self, 'slug', None)
        if not slug:
            return 'http://placehold.it/55x55'
        return static(
            'icons/icons_achievements/{slug}.png'.format(slug=slug.replace('-', '_')))


@register.assignment_tag
def most_awarded_badges(client_id=None, limit=5):
    """
    Usage:

        {% load badges_tags %}

        {% most_awarded_badges client_id=user.client_id limit=5 as badges %}

        {% for badge in badges %}
            Name: {{badge.name}}
            ...
        {% endfor%}
    """

    qs = BadgeAward.objects
    user_qs = User.objects
    if client_id:
        qs = qs.filter(user__client_id=client_id)
        user_qs = user_qs.filter(client_id=client_id)

    users_count = user_qs.aggregate(count=Count('*'))['count']

    qs = qs \
        .values_list('slug', 'level') \
        .annotate(count=Count('id')) \
        .order_by('-count', 'level', 'slug')[:limit]

    def lazy():
        return [
            MostAwardedBadge(
                detail=badges._registry[slug].levels[level],
                count=count,
                slug=slug,
                users_count=users_count
            )
            for slug, level, count in qs
        ]

    return SimpleLazyObject(lazy)



from django.shortcuts import render
from django.contrib.auth.decorators import login_required


@login_required
def awards_list(request):
    context = {}
    return render(request, 'awards/awards_list.html', context)



from .internals import badges
from .base import Badge, BadgeDetail, BadgeAwarded  # NOQA

__version__ = '0.3.0'

default_app_config = 'awards.apps.AwardsConfig'


possibly_award_badge = badges.possibly_award_badge



from celery.task import Task

from ..notifications.contextmanagers import BatchNotifications


class AsyncBadgeAward(Task):
    ignore_result = True

    def run(self, badge, state, **kwargs):
        # from celery.contrib import rdb; rdb.set_trace()
        with BatchNotifications():
            badge.actually_possibly_award(**state)




# coding: utf-8

from .base import Badge


class BadgeCache(object):
    """
    This is responsible for storing all badges that have been registered, as
    well as providing the pulic API for awarding badges.

    This class should not be instantiated multiple times, if you do it's your
    fault when things break, and you get to pick up all the pieces.
    """
    def __init__(self):
        self._event_registry = {}
        self._registry = {}

    def register(self, badge_cls):
        # We should probably duck-type this, but for now it's a decent sanity
        # check.
        assert issubclass(badge_cls, Badge)
        badge = badge_cls()
        if badge.slug in self._registry:
            raise ValueError("{} is already registered.".format(badge_cls))
        self._registry[badge.slug] = badge
        for event in badge.events:
            self._event_registry.setdefault(event, []).append(badge)
        return badge_cls

    def possibly_award_badge(self, event, callback=None, **state):
        for badge in self._event_registry[event]:
            awarded = badge.possibly_award(**state)
            if awarded and callback:
                callback(awarded)


badges = BadgeCache()



# coding: utf-8
from __future__ import unicode_literals

from datetime import datetime

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.utils.encoding import python_2_unicode_compatible
from django.utils.text import force_text

from . import settings as app_settings


@python_2_unicode_compatible
class BadgeAward(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="badges_earned")
    slug = models.CharField(_('slug'), max_length=255)
    level = models.IntegerField(_('level'))

    points = models.PositiveIntegerField(_('points'), default=0)

    awarded_on = models.DateTimeField(_('awarded on'), default=datetime.now)
    created_on = models.DateTimeField(_('created on'), auto_now_add=True)

    class Meta:
        verbose_name = _('badge award')
        verbose_name_plural = _('badge awards')

    def __getattr__(self, attr):
        return getattr(self._badge, attr)

    def __str__(self):
        return self.name

    def __repr__(self):
        return u'<%s: %s (%s) awarded to %s>' % (
            self.__class__.__name__,
            self.name, self.level, self.user)

    @property
    def badge(self):
        return self

    @property
    def _badge(self):
        from .internals import badges
        return badges._registry[self.slug]

    @property
    def name(self):
        return force_text(self._badge.levels[self.level].name)

    @property
    def description(self):
        return force_text(self._badge.levels[self.level].description)

    @property
    def info(self):
        return self._badge.levels[self.level]

    @property
    def image_url(self):
        slug = getattr(self.info, 'slug', self.slug)
        return static(
            app_settings.IMAGE_URL.format(slug=slug.replace('-', '_')))

    @property
    def progress(self):
        return self._badge.progress(self.user, self.level)

    @staticmethod
    def get_user_points(user_id):
        return BadgeAward.objects \
            .filter(user_id=user_id) \
            .aggregate(points=models.Sum('points')) \
            .get('points', 0)



# coding: utf-8

from rest_framework import serializers

from .models import BadgeAward


class BadgesSerializer(serializers.ModelSerializer):
    image_url = serializers.SerializerMethodField()
    awarded_on = serializers.DateTimeField(read_only=True)

    class Meta:
        model = BadgeAward
        fields = ('slug', 'level', 'points', 'awarded_on', 'name', 'description', 'image_url')

    def get_image_url(self, obj):
        return self.context['view'].request.build_absolute_uri(obj.image_url)



# coding: utf-8

from django.conf import settings
from django.utils.translation import ugettext as _
from django.core.urlresolvers import reverse

if "notify" in settings.INSTALLED_APPS:
    from notify.signals import notify as notification
else:
    notification = None


def notify_awarded(badge):
    if not notification:
        return

    verb = _("awarded a badge")
    nf_type = "badge_awarded"

    notification.send(
        badge.user,
        recipient=badge.user,
        actor=badge.user,
        verb=verb,
        target=badge,
        nf_type=nf_type,
        target_url=reverse('awards:list')
    )




from .signals import badge_awarded


class BadgeAwarded(object):
    def __init__(self, level=None, user=None):
        self.level = level
        self.user = user


class BadgeDetail(object):
    def __init__(self, name=None, description=None, **kwargs):
        self.name = name
        self.description = description
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return unicode(self.name)


class Badge(object):
    async = False
    multiple = False

    def __init__(self):
        assert not (self.multiple and len(self.levels) > 1)
        for i, level in enumerate(self.levels):
            if not isinstance(level, BadgeDetail):
                self.levels[i] = BadgeDetail(level)

    def possibly_award(self, **state):
        """
        Will see if the user should be awarded a badge.  If this badge is
        asynchronous it just queues up the badge awarding.
        """
        assert "user" in state

        try:
            user_id = state['user'].pk
        except AttributeError:
            user_id = state['user']

        state['user_id'] = user_id

        if self.async:
            from .tasks import AsyncBadgeAward
            if 'user' in state:
                del state['user']
            state = self.freeze(**state)
            extra = {
                "countdown": getattr(self, 'countdown', None),
            }
            AsyncBadgeAward.apply_async((self, state), **extra)
            return
        return self.actually_possibly_award(**state)

    def actually_possibly_award(self, **state):
        """
        Does the actual work of possibly awarding a badge.
        """
        from .models import BadgeAward
        from .notification import notify_awarded

        user_id = state["user_id"]
        force_timestamp = state.pop("force_timestamp", None)
        awarded = self.award(**state)
        if awarded is None:
            return
        if awarded.user is not None:
            user_id = awarded.user.pk
        if awarded.level is None:
            assert len(self.levels) == 1
            awarded.level = 0
        awarded = awarded.level
        assert awarded < len(self.levels)
        if (not self.multiple and
                BadgeAward.objects.filter(user_id=user_id, slug=self.slug, level=awarded)):
            return
        extra_kwargs = {
            'points': getattr(self.levels[awarded], 'points', 0)
        }
        if force_timestamp is not None:
            extra_kwargs["awarded_at"] = force_timestamp
        badge = BadgeAward.objects.create(
            user_id=user_id, slug=self.slug,
            level=awarded, **extra_kwargs)
        if 'user' in state:
            badge.user = state['user']
        badge_awarded.send(sender=self, badge_award=badge)
        notify_awarded(badge)
        return badge

    def freeze(self, **state):
        return state



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _
from django.utils.module_loading import autodiscover_modules


class AwardsConfig(AppConfig):
    name = 'awards'
    verbose_name = _("Awards")

    def ready(self):
        autodiscover_modules('badges')
        autodiscover_modules('badges_receivers')
